require "singleton"

# primitive constants
PRIM_POINT = 0
PRIM_LINE = 1
PRIM_LINESTRIP = 2
PRIM_TRIANGLE = 3
PRIM_TRIANGLESTRIP = 4
PRIM_TRIANGLEFAN = 5
PRIM_SPRITE = 6
PRIM_GOURAUD = 8
PRIM_TEXTURE = 16
PRIM_FOGGING = 32
PRIM_ALPHA = 64
PRIM_AA = 128
PRIM_UV = 256
PRIM_CONTEXT1 = 512
PRIM_FIX = 1024

# test register constants
ANEVER = 0
AALWAYS = 1
ALESS = 2
ALEQUAL = 3
AEQUAL = 4
AGEQUAL = 5
AGREATER = 6
ANOTEQUAL = 7
KEEP = 0
FB_ONLY = 1
ZB_ONLY = 2
RGB_ONLY = 3
ZNEVER = 0
ZALWAYS = 1
ZGEQUAL = 2
ZGREATER = 3

# pixel storage constants
PSMCT32 = 0
PSMCT24 = 1
PSMCT16 = 2
PSMCT16S = 10
PSMT8 = 19
PSMT4 = 20
PSMT8H = 27
PSMT4HL = 36
PSMT4HH = 44
PSMZ32 = 48
PSMZ24 = 49
PSMZ16 = 50
PSMZ16S = 58

# texture function constants
MODULATE = 0
DECAL = 1
HIGHLIGHT = 2
HIGHLIGHT2 = 3

# filtering constants
NEAREST = 0
LINEAR = 1

# to_bin functions for common datatypes, used by VuData.add
class String
	def to_bin
		self
	end
end

class Integer
	def to_bin
		[self].pack("l")
	end
end

class Float
	def to_bin
		[self].pack("f")
	end
end

class VuData
	include Singleton

	def initialize
		@stream = ""
		@labels = []
	end

	def add(data)
		@stream += data.to_bin
	end

	def VuData.add(data)
		VuData.instance.add(data)
	end

	def addLightMatrix(v0, v1, v2, v3)
		('x'..'z').each do |f|
			[v0, v1, v2, v3].each do |v|
				v.normalise!
				add(v.send f)
			end
		end
	end
	
	def VuData.addLightMatrix(v0, v1, v2, v3)
		VuData.instance.addLightMatrix(v0, v1, v2, v3)
	end

	def addLabel(label)
		raise "Stream position not divisible by 16" if (@stream.size & 15) != 0
		@labels.push([label, @stream.size / 16])
	end

	def VuData.addLabel(label)
		VuData.instance.addLabel(label)
	end

	def getLabel(label)
		@labels.assoc(label)[1]
	end

	def VuData.getLabel(label)
		VuData.instance.getLabel(label)
	end

	def setSize(size)
		@stream = @stream[0, size * 16]
	end

	def VuData.setSize(size)
		VuData.instance.setSize(size)
	end

	def align
		@stream += 0.chr while (@stream.size & 15) != 0
	end
	
	def VuData.align
		VuData.instance.align
	end

	def write
		raise "Stream size not divisible by 16" if (@stream.size & 15) != 0
		File.open(ARGV[0]+".i", "w") do |f|
			@labels.each {|l| f.printf("\t.equ\t\t%s, 0x%x\n", l[0], l[1]) }
			f.printf("\t.include\t\"%s.vu\"\n", ARGV[0])
		end
		File.open(ARGV[0]+".dat", "wb") {|f| f.print(@stream) }
		lm = {}
		@labels.each {|l| lm[l[1]] = l[0]}
		File.open(ARGV[0]+".ds", "w") do |f|
			(@stream.size / 16).times do |i|
				f.puts lm[i]+":" if lm.has_key?(i)
				f.print "\t"
				@stream[i * 16, 16].reverse.each_byte {|b| f.printf "%02x ", b}
				f.print "\n"
			end
		end
	end

	def VuData.write
		VuData.instance.write
	end
end

class GifPacket
	attr_accessor :nloop, :eop, :prim, :flg, :nreg, :regs
	
	def initialize(prim = nil, nreg = -1, nloop = -1, regs = [])
		@nloop = nloop
		@eop = 1
		@prim = prim
		@flg = 0
		@nreg = nreg
		@regs = regs
		@packet = ""
	end

	def GifPacket.build(*args)
		gp = GifPacket.new(*args)
		yield gp if block_given?
		gp.dump
	end

	def addReg(reg)
  		regs.push(reg)
	end

	def getLong(val)
		[val & (2**32 - 1), val >> 32].pack("LL")
	end

	def addAD(address, data)
		addReg(0xe)
		@packet += getLong(data) + getLong(address)
	end

	def addAlpha(a, b, c, d, fix, ctxt = 0)
		addAD(0x42 + ctxt, a + (b << 2) + (c << 4) + (d << 6) + (fix << 32))
	end

	def addClamp(wms, wmt, minu = 0, maxu = 0, minv = 0, maxv = 0, ctxt = 0)
		addAD(0x8 + ctxt, wms + (wmt << 2) + (minu << 4) + (maxu << 14) + (minv << 24) + (maxv << 34))
	end

	def addFrame(fbp, fbw, psm, fbmsk = 0, ctxt = 0)
		addAD(0x4c + ctxt, fbp + (fbw << 16) + (psm << 24) + (fbmsk << 32))
	end

	def addTest(ate, atst, aref, afail, date, datm, zte, ztst, ctxt = 0)
		addAD(0x47 + ctxt, ate + (atst << 1) + (aref << 4) + (afail << 12) +
			(date << 14) + (datm << 15) + (zte << 16) + (ztst << 17))
	end

	def addTex0(tbp0, tbw, psm, tw, th, tcc, tfx, cbp = 0, cpsm = 0,
		csm = 0, csa = 0, cld = 0, ctxt = 0)
		addAD(0x6 + ctxt, tbp0 + (tbw << 14) + (psm << 20) +
			(tw << 26) + (th << 30) + (tcc << 34) + (tfx << 35) +
			(cbp << 37) + (cpsm << 51) + (csm << 55) +
			(csa << 56) + (cld << 61))
	end

	def addTex1(lcm, mxl, mmag, mmin, mtba, l, k, ctxt = 0)
		addAD(0x14 + ctxt, lcm + (mxl << 2) + (mmag << 5) + (mmin << 6) +
			(mtba << 9) + (l << 19) + (k << 32))
	end

	def addXyoffset(x, y, ctxt = 0)
		addAD(0x18 + ctxt, (x * 16).to_i + ((y * 16).to_i << 32))
	end

	def addRgbaq(r, g, b, a)
		addReg(1)
		@packet += [r, g, b, a].pack("L*")
  	end

	def addUv(u, v)
		addReg(3)
		@packet += [u * 16, v * 16, 0, 0].pack("L*")
	end

	def addXyz2(x, y, z, adc = 0)
  		addReg(5)
  		@packet += [x*16, y*16, z, adc << 15].pack("L*")
  	end

	def dump
		if @nreg == -1
			@regs.size.downto(1) do |i|
				next unless (@regs.size % i) == 0
				f = true
				i.upto(@regs.size - 1) do |j|
					f = false if @regs[j] != @regs[j % i]
				end
				@nreg = i if f
			end
		end
		if @nloop == -1
			raise "Number of set registers is no divisible by nreg" if (@regs.size % @nreg) != 0
			@nloop = @regs.size / @nreg
		end
		if @prim != nil
			@pre = 1
		else
			@prim = @pre = 0
		end
		regs = ""
		@nreg.times {|i| regs += "%x" % @regs[i]}
		regs = getLong(regs.reverse.hex)
		output = getLong(@nloop + (@eop << 15) + (@pre << 46) +
			(@prim << 47) + (@flg << 58) + (@nreg << 60)) + regs + @packet
		VuData.add(output)
	end
end

class Vector
	attr_accessor :x, :y, :z, :w
	
	def initialize(x = 0.0, y = 0.0, z = 0.0, w = 0.0)
		@x = x.to_f
		@y = y.to_f
		@z = z.to_f
		@w = w.to_f
	end

	def *(o)
		if o.kind_of?(Vector)
			@x * o.x + @y * o.y + @z * o.z + @w * o.w
		else
			Vector.new(@x * o, @y * o, @z * o, @w * o)
		end
	end

	def +(o)
		Vector.new(@x + o.x, @y + o.y, @z + o.z, @w + o.w)
	end

	def -(o)
		Vector.new(@x - o.x, @y - o.y, @z - o.z, @w - o.w)
	end

	def cross(o)
		Vector.new(@y * o.z - @z * o.y, @z * v.x - @x * v.z,
			@x * v.y - @y * v.x)
	end

	def sqrLength
		self * self
	end

	def length
		Math.sqrt(sqrLength)
	end

	def normalise
		self * (1.0 / length)
	end

	def normalise!
		f = 1.0 / length
		@x *= f
		@y *= f
		@z *= f
		@w *= f
		self
	end

	def to_bin
		[@x, @y, @z, @w].pack("f*")
	end
end

def Vector(*args)
  Vector.new(*args)
end

class Matrix
	attr_accessor :x, :y, :z, :pos

	def initialize
		@x = Vector.new
		@y = Vector.new
		@z = Vector.new
		@pos = Vector.new
	end

	def to_bin
		@x.to_bin + @y.to_bin + @z.to_bin + @pos.to_bin
	end
end
