# op constants
NOISE = 0
BLUR = 1
EMBOSS = 2
ADJUST = 3
ADD = 4
GREY = 5
MUL = 6
INVERT = 7
BOX = 8
ROTZOOM = 9

# noise(red blue green color)
# blur(buffer x y mul add)
# emboss(buffer)
# adjust(buffer mul add times)
# add(buffer buffer mul mul)
# grey(buffer red green blue)
# mul(buffer buffer brightness)
# invert(buffer)
# box(buffer? x0 y0 x1 y1 red green blue)
# rotzoom(buffer xoffset yoffset rotate zoom)

include Math

def floatInt(f, i)
	[i].pack("s") + [f].pack("f")[2,2]
end

# treenode class
class Treenode
	attr_reader :value, :children

	# init & parse a script list
	def initialize(list)
		raise "missing ')'" if list.empty?
		@children = []
		@value = list.shift
		return if list[0] != "("
		list.shift
		@children.push(Treenode.new(list)) while list[0] != ")"
		list.shift
	end

	# convert treenode and its children to text
	def to_s
		@value + ("(" + @children.join(" ") + ")" unless @children.empty?).to_s
	end

	def eval
		begin
			Kernel::eval(to_s)
		rescue
			raise "couldn't evaluate: "+to_s
		end
	end

	# count the references to a given buffer
	def count(buffer)
		if @value == "load"
			buffer.reference if @children[0].value == buffer.name
		else
			@children.each {|c| c.count(buffer)}
		end
	end

	# compile treenode
	def compile
		case @value
			when "load"
				return "", Buffer.get(@children[0].value)
			when "noise"
				buf = Buffer.new
				return floatInt(@children[0].eval * 255, NOISE) +
					floatInt(@children[1].eval * 255, buf.index) +
					[@children[2].eval * 255, @children[3].eval].pack("f*"), buf
			when "blur"
				code, inbuf = @children[0].compile
				inbuf.destroy
				obuf = Buffer.new
				tbuf = Buffer.new
				tbuf.destroy
				return code + floatInt(@children[4].eval * 128, BLUR) +
					floatInt(@children[1].eval, inbuf.index) +
					floatInt(@children[2].eval, obuf.index) +
					floatInt(@children[3].eval * 128, tbuf.index), obuf
			when "emboss"
				code, inbuf = @children[0].compile
				obuf = Buffer.new
				inbuf.destroy
				return code + [EMBOSS, inbuf.index, obuf.index, 0].pack("L*"), obuf
			when "adjust"
				code, inbuf = @children[0].compile
				obuf = Buffer.new
				inbuf.destroy
				obuf2 = Buffer.new
				count = @children[3].eval
				(((count & 1) == 0) ? obuf : obuf2).destroy
				return code + floatInt(@children[1].eval * 128, ADJUST) +
					floatInt(@children[2].eval * 128, inbuf.index) +
					[obuf.index].pack("L") + floatInt(count, obuf2.index),
					((count & 1) != 0) ? obuf : obuf2
			when "add"
				code1, inbuf1 = @children[0].compile
				code2, inbuf2 = @children[1].compile
				inbuf1.destroy
				obuf = Buffer.new
				inbuf2.destroy
				return code1 + code2 + floatInt(@children[2].eval * 128, ADD) +
					floatInt(@children[3].eval * 128, inbuf1.index) +
					[inbuf2.index, obuf.index].pack("L*"), obuf
			when "grey"
				code, inbuf = @children[0].compile
				obuf = Buffer.new
				inbuf.destroy
				return code + floatInt(@children[1].eval * 128, GREY) +
					floatInt(@children[2].eval * 128, inbuf.index) +
					floatInt(@children[3].eval * 128, obuf.index) +
					[0].pack("L*"), obuf
			when "mul"
				code1, inbuf1 = @children[0].compile
				code2, inbuf2 = @children[1].compile
				inbuf1.destroy
				obuf = Buffer.new
				inbuf2.destroy
				return code1 + code2 + floatInt(@children[2].eval * 128, MUL) +
					[inbuf1.index, inbuf2.index, obuf.index].pack("L*"), obuf
			when "invert"
				code, inbuf = @children[0].compile
				obuf = Buffer.new
				inbuf.destroy
				return code + [INVERT, inbuf.index, obuf.index, 0].pack("L*"), obuf
			when "box"
				if @children.size == 8
					code, inbuf = @children.shift.compile
				else
					code = ""
					inbuf = nil
				end
				obuf = Buffer.new
				inbuf.destroy if inbuf
				return code + floatInt(@children[0].eval * 127.99, BOX) +
					floatInt(@children[1].eval * 127.99, obuf.index +
						(inbuf ? inbuf.index : 0) * 256) +
					floatInt(@children[2].eval * 127.99, @children[4].eval * 255 +
						(@children[5].eval * 255).to_i * 256) +
					floatInt(@children[3].eval * 127.99, @children[6].eval * 255),
					obuf
			when "rotzoom"
				code, inbuf = @children[0].compile
				obuf = Buffer.new
				inbuf.destroy
				return code + floatInt(@children[1].eval, ROTZOOM) +
					floatInt(@children[2].eval, inbuf.index) +
					floatInt(@children[3].eval, obuf.index) +
					floatInt(@children[4].eval, 0), obuf
			else
				raise "Unknown op '"+@value+"'"
		end
	end
end

class Buffer
	attr_reader :name, :index, :refcount

	@@bufstack = (35...64).to_a.reverse
	@@symbols = {}

	def initialize
		raise "No free buffer" if @@bufstack.empty?
		@index = @@bufstack.pop
		@refcount = 1
	end

	def insert(name)
		@name = name
		@@symbols[name] = self
	end

	def reference
		@refcount += 1
		self
	end
	
	def destroy
		@refcount -= 1
		return if @refcount != 0
		@@bufstack.push(@index)
  		@@symbols[@name] = nil if @name
	end

	def Buffer.get(name)
		raise "Unknown buffer "+name unless @@symbols.has_key?(name)
		@@symbols[name]
	end

	def Buffer.cleanup
	end

	def Buffer.writeSymbols(filename)
		File.open(filename, "w") do |f|
			@@symbols.each {|s, v| f.puts "#{s} #{v.index}" if v}
		end
	end
end

# read file
s = ""
File.foreach(ARGV[0]) {|l| s += l if l.strip[0, 1] != "#"}

# tokenise (verrry simply)
s = s.gsub(/\(/, " ( ").gsub(/\)/, " ) ").split(" ")

# build a list of trees
treelist = []
treelist.push(Treenode.new(s)) until s.empty?
puts treelist

stream = ""

for tree in treelist
	next if tree.value[0,1] == "-"
	code, buffer = tree.children[1].compile
	buffer.insert(tree.children[0].value)
	treelist.each {|t| t.count(buffer) if t.value[0].chr != "-"}
	puts "#{tree.children[0]}: #{buffer.index}"
	buffer.destroy if tree.value == "store"
	stream += code if buffer.refcount > 0
end

Buffer.cleanup

stream += [-1, 0, 0, 0].pack("l*")

File.open(ARGV[1], "wb") {|f| f.print stream}

Buffer.writeSymbols("texturesymbols.txt")
